#ifndef __CELEM_H__
#define __CELEM_H__

#include <vector>

#define ELEM_GROUP_SINGLE 0

// types of vector elements
enum
 {
	ELEM_TYPE_LINES,
	/*ELEM_TYPE_MULTILINE,*/
	/*ELEM_TYPE_TRIANGLE,*/
	/*ELEM_TYPE_RECTANGLE,*/
	ELEM_TYPE_ELLIPSE,
	ELEM_TYPE_ARC,
	ELEM_TYPE_POLYGON,
	ELEM_TYPE_BEZIER,
	ELEM_TYPE_TEXT
 };


/********************************************/
/* Abstract parent class of vector elements */
/********************************************/
class CElem
 {
  public:
	struct tPoint                                                       // geometric point structure (single item)
	 {
		double x;
		double y;
	 };
	struct tPointLink                                                   // geometric point structure (item of linked list)
	 {
		double x;
		double y;
		struct tPointLink *next;
	 } *points;                                                         // gives syntax sense to some virtual functions 

	int elemtype;                                                       // type of element (ELEMENT_LINE | ELEMENT_TRIANGLE...)

	struct tCommon                                                      // common atributes, saved at once
	 {
		unsigned pointcnt;                                              // number of points this element has (=length of linked list)
		int depth;
		bool selected;
		bool visible;
		int groupId;                                                    // ID of group in which this element belongs (0 for nonea)
		unsigned short colorR;
		unsigned short colorG;
		unsigned short colorB;
		unsigned short colorRbg;
		unsigned short colorGbg;
		unsigned short colorBbg;
		int lineWidth;
		int lineStyle;
		bool filled;
	 } common;

	CElem();
	virtual ~CElem();
	virtual void draw() const = 0;
	virtual struct tPoint *boundary( unsigned *ptcnt ) const = 0;       // get array of points which form boundary box (newed)
	virtual double distance( struct tPoint click ) const = 0;           // get shortest distance^2 between element and click (world coords)
	// whether element is located inside rectangular area and either whole (all points are there) or one point is enough
	virtual bool inside( struct tPoint pt1, struct tPoint pt2, bool whole ) const = 0;
	void addPoint ( unsigned pt, double newX, double newY );            // add new point to list before pt (too high pt adds at the end)
	bool delPoint ( unsigned pt );
	bool movePoint( unsigned pt, double newX, double newY );            // pt is point's order in linked list of points
	bool getPoint ( unsigned pt, double *oldX, double *oldY );          // return value is false, when coord are invalid (pt >= list length)
	void *save( int *len ) const;                                       // generate saving data; len gives record bytecount
	int  load( const void *ptr );                                       // load element from memory, returns read bytecount

	static unsigned elemcnt;

  protected:
	void allocPoints();                                                 // allocate linked list of points (add pointcnt items)
	void deallocPoints();
	void saveCommon( void *ptr ) const;                                 // save common atributes and TYPE of element on [ptr]
	void loadCommon( const void *ptr );                                 // load common atributes and TYPE of element on [ptr]

  public:
	// draw vector (depth-ordered) of elements, with regard to visibility and selected groups (boundary box is drawn once for whole group)
	static void drawElems(  std::vector<CElem *> &elems );
	// select element(s) near (tolerance) to (world) coords (x,y) and return number of new selected items (additive doesn't unselect others
	static int selectElems( std::vector<CElem *> &elems, double x, double y, bool additive, double tolerance );
	// select element(s) laying in rectangular area (world coords) and return number of new selected items ('whole' as in instance method)
	static int selectElemsInside( std::vector<CElem *> &elems, double x1, double y1, double x2, double y2, bool additive, bool whole );
	// group all elements which are selected and return number of grouped ones
	static int groupElems( std::vector<CElem *> &elems );
	// ungroup all elements which are selected and return number of ungrouped ones
	static int ungroupElems( std::vector<CElem *> &elems );
	// move elements which are selected (move coords are world relative)
	static int moveElems( std::vector<CElem *> &elems, double x, double y );
	// delete elements which are selected
	static int delElems( std::vector<CElem *> &elems );

	//TODO: implement
	// update common attributes of objects
	static int updateElems ( std::vector<CElem *> &elems );
 };

/********************************************/
/* Generic line element (single or multi)   */
/********************************************/
class CElemLines : public CElem
 {
  public:
	CElemLines( unsigned pointcnt );                                    // pointcnt is number of points line has (3+ for multiline)
	virtual ~CElemLines();
	void draw() const;
	struct tPoint *boundary( unsigned *ptcnt ) const;
	double distance( struct tPoint click ) const;
	bool inside( struct tPoint pt1, struct tPoint pt2, bool whole ) const;
 };

/********************************************/
/* Ellipse element                          */
/********************************************/
class CElemEllipse : public CElem
 {
  public:
	CElemEllipse();
	virtual ~CElemEllipse();
	void draw() const;
	struct tPoint *boundary( unsigned *ptcnt ) const;
	double distance( struct tPoint click ) const;
	bool inside( struct tPoint pt1, struct tPoint pt2, bool whole ) const;
 };

/********************************************/
/* Polygon element                          */
/********************************************/
class CElemPolygon : public CElem
 {
  public:
	CElemPolygon( unsigned linecnt );                                   // linecnt is number of lines in polygon border
	virtual ~CElemPolygon();
	void draw() const;
	struct tPoint *boundary( unsigned *ptcnt ) const;
	double distance( struct tPoint click ) const;
	bool inside( struct tPoint pt1, struct tPoint pt2, bool whole ) const;
 };

/********************************************/
/* Text (label) element                     */
/********************************************/
//TODO
class CElemText : public CElem
 {
  public:
	CElemText();
	virtual ~CElemText();
	void draw() const;
	struct tPoint *boundary( unsigned *ptcnt ) const;
	double distance( struct tPoint click ) const;
	bool inside( struct tPoint pt1, struct tPoint pt2, bool whole ) const;
 };


#endif

